"""
This file is part of VexilPy (elemenom/vexilpy).

VexilPy is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

VexilPy is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VexilPy. If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Optional
from ..safety.handler import handle

from ..app.appobject import AppObject
from ..app.app import WebApp

class TagObject(AppObject):
    @handle
    def __init__(self, name: str, tag: str, app: WebApp, args: Optional[WebApp] = None) -> None:
        super().__init__(name, app)
        self.tagstr: str = tag
        self.singular(f"<{self.tagstr} {args or ''}>")

    @handle
    def on_run(self) -> None:
        ...

    @handle
    def __exit__(self, *_) -> None:
        self.singular(f"</{self.tagstr}>")


"""
This file is part of VexilPy (elemenom/vexilpy).

VexilPy is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

VexilPy is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VexilPy. If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Optional
from ..safety.handler import handle

from ..app.app import WebApp
from ..server.json import JsonServer

class JsonWebApp(WebApp):
    @handle
    def __init__(self, path: Optional[str] = None) -> None:
        super().__init__(JsonServer(path or "index.json"))

@handle
def jsonapp(path: Optional[str] = None) -> JsonWebApp:
    return JsonWebApp(path)


"""
This file is part of VexilPy (elemenom/vexilpy).

VexilPy is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

VexilPy is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VexilPy. If not, see <https://www.gnu.org/licenses/>.
"""

import sys
from ..vexilpy.pwsh import pwsh

match sys.argv[1]:
    case "visit":
        match sys.argv[2]:
            case "vexilpy-github":
                pwsh(f"start https://github.com/elemenom/vexilpy/")

            case "vexilpy-pypi":
                pwsh(f"start https://pypi.org/project/vexilpy/")

            case "vexilpy-github-issues":
                pwsh(f"start https://github.com/elemenom/vexilpy/issues/")

            case "vexilpy-github-pr":
                pwsh(f"start https://github.com/elemenom/vexilpy/pulls/")

            case "vexilpy-github-pr-new":
                pwsh(f"start https://github.com/elemenom/vexilpy/compare/")

            case "vexilpy-github-insights":
                pwsh(f"start https://github.com/elemenom/vexilpy/pulse/")

            case "vexilpy-documentation":
                pwsh(f"start https://github.com/elemenom/vexilpy?tab=readme-ov-file#welcome-to-vexilpy/")

            case "vexilpy-license":
                pwsh(f"start https://github.com/elemenom/vexilpy/blob/main/LICENSE")

            case "author-github":
                pwsh(f"start https://github.com/elemenom/")

            case "author-pypi":
                pwsh(f"start https://pypi.org/user/elemenom/")

            case "vexilpy-github-download":
                pwsh(f"start https://github.com/elemenom/vexilpy/releases/")

    case "upgrade":
        pwsh("pip install vexilpy --upgrade")

    case "install":
        pwsh("pip install vexilpy")

    case "install-version":
        pwsh(f"pip install vexilpy=={sys.argv[2]}")

    case "info":
        pwsh("pip show vexilpy")

    case "uninstall":
        pwsh("pip uninstall vexilpy")

    case "documentation":
        import vexilpy

        print(vexilpy.__doc__)

    case "feedback":
        pwsh(f"start https://github.com/elemenom/vexilpy/issues/new/")

    case "license":
        print(__doc__)


"""
This file is part of VexilPy (elemenom/vexilpy).

VexilPy is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

VexilPy is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VexilPy. If not, see <https://www.gnu.org/licenses/>.
"""

import os

from typing import Optional, Any
from ..safety.handler import handle

from ..safety.logger import logger
from ..launcher.launch import launch
from ..app.supportswith import SupportsWithKeyword
from ..app.supportedtags import supported_tags
from ..script.ctrl import CTRLScript
from ..style.style import StyledAppAttachment

class AppObject(SupportsWithKeyword):
    @handle
    def __init__(self, name: Optional[str] = None, app: Any = None) -> None:
        super().__init__()

        from ..app.app import app as app_

        if app is None:
            logger().fatal("Please provide a parent app.")
            exit(1)

        self.server: Any = app.server
        self.name: str = name
        self.app: app_ = app
        self.style_: StyledAppAttachment | None = app.style
        self.new_stylesheet_path: str = f"{self.name}.css"
        self.ctrl: CTRLScript | None = None

        self.init_supported_tags(supported_tags)
        self.init_ctrlscript()

        self.on_run()

    @handle
    def on_run(self) -> None:
        logger().info("Please wait while we build your HTML file for you.")

        self.singular("<!DOCTYPE html>")
        self.singular("<html>")
        self.singular(f"<link rel=\"stylesheet\" href=\"{self.new_stylesheet_path}\">")
        self.singular(f"<script src=\"{self.name}.js\"></script>")

    @handle
    def init_ctrlscript(self) -> None:
        self.ctrl = CTRLScript(self.name)

    @handle
    def init_supported_tags(self, supported_tags_: list[str]) -> None:
        for tag in supported_tags_:
            exec(f"self.{tag} = lambda **args: self.tag({repr(tag)}, **args)", {"self": self})

    @handle
    def tag(self, type_: str, **kwargs: str | int) -> Any:
        from ..app.tagobject import TagObject

        final_kwargs = ""

        for key, value in list(kwargs.items()):
            final_kwargs += f"{key}=\"{value}\" "

        return TagObject(self.name, type_, self.app, final_kwargs)

    @handle
    def singular(self, ln: str) -> None:
        with open(f"{self.name}.html", "a") as file:
            file.write(ln + "\n")

    @handle
    def __exit__(self, *_) -> None:
        self.exit()

    @handle
    def exit(self) -> None:
        logger().info("Building has been finished successfully.")

    @handle
    def __getattr__(self, item) -> None:
        logger().error(f"Attribute '{item}' for 'SelfApp' does not exist. Skipping to next line.")

    @handle
    def pass_to_server(self) -> None:
        self.singular("</html>")

        if self.server is None:
            logger().fatal("Cannot pass vexilpy app script to server when no server was provided.")
            exit(1)

        logger().info(f"Passed '{self.name}' vexilpy app script to host '{type(self.server).__name__}'")

        if self.style is not None:
            logger().info(f"Passed stylesheet '{self.new_stylesheet_path}' to pebl script '{self.name}'.")

            self.style_.close()

        launch(self.server)

        logger().info("Continuing in pebl app to clear cache.")

        try: os.remove(f"{self.name}.html")
        except FileNotFoundError:
            logger().warn(f"Could not find '{self.name}', so it cannot be cleaned.")

        try: os.remove(self.ctrl.path)
        except FileNotFoundError:
            logger().warn(f"Could not find '{self.ctrl.path}', so it cannot be cleaned.")

        if self.style_ is not None:
            try: os.remove(self.new_stylesheet_path)
            except FileNotFoundError:
                logger().warn(f"Could not find '{self.new_stylesheet_path}', so it cannot be cleaned.")


"""
This file is part of VexilPy (elemenom/vexilpy).

VexilPy is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

VexilPy is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VexilPy. If not, see <https://www.gnu.org/licenses/>.
"""

import sys
from functools import wraps
from typing import Callable, Any

def handle(fn: Callable) -> Callable:
    @wraps(fn)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        from ..safety.logger import logger

        from ..safety.safe_input import safe_input

        try:
            result: Any = fn(*args, **kwargs)
            return result

        except KeyboardInterrupt:
            try:
                logger().warning("Abandoning VexilPy while it's still running may cause currently open\n"
                               "servers to continue to take up their address in your memory.\n"
                               "Please exit the program safely or press CTRL+C/CTRL+D again.\n"
                               "Press ENTER to cancel this exit attempt."
                )
                safe_input()

            except KeyboardInterrupt:
                logger().fatal("Exiting program... KeyboardInterrupt")

                exit(0)

            return

        except Exception as err:
            from vexilpy import verbose

            if verbose:
                logger().fatal(f"An error was caught but the handler is offline. Recovery process aborted.\n"
                               f"hint: this might be because the current running VexilPy instance is Verbose.")

                raise err

            else:

                logger().error(f"VexilPy encountered an error `{type(err).__name__}` in function `{fn.__name__}`,\n"
                               f"but the system managed to recover.\n"
                               f"Use `python {" ".join(sys.argv)} --Verbose` to disable error recovery.")

                return

    return wrapper


"""
This file is part of VexilPy (elemenom/vexilpy).

VexilPy is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

VexilPy is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VexilPy. If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Any, Callable
from ..safety.handler import handle

from ..app.appobject import AppObject
from ..app.app import WebApp
from ..app.supportswith import SupportsWithKeyword

class StandardAppExportObject(SupportsWithKeyword):
    @handle
    def __init__(self, app_: WebApp, object_: Any = AppObject, *args: Any, **kwargs: Any) -> None:
        self.app: WebApp = app_
        self.name: str = "index" if self.app.fn.__name__.lower() == "base" else self.app.fn.__name__
        self.object: type[object_] = object_
        self.args: tuple[Any, ...] = args
        self.kwargs: dict[str, Any] = kwargs
        self.postb: Any = None

    @handle
    def open(self) -> Any:
        app_object: Any = self.object(self.name, self.app)
        app_object.exit()

        try: return handle(self.app.fn)(app_object, *self.args, **self.kwargs)
        finally: app_object.pass_to_server()

    @handle
    def init_postbuilding(self, type_: Callable) -> None:
        self.postb = type_(self.name, self.app)

    @handle
    def get_object_type(self) -> Any:
        return self.object

    @handle
    def get_app(self) -> WebApp:
        return self.app


"""
This file is part of VexilPy (elemenom/vexilpy).

VexilPy is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

VexilPy is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VexilPy. If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Any

from ..safety.uselessobject import UselessObject
from ..safety.handler import handle

logger_literal: Any = NotImplementedError

@handle
def logger() -> Any:
    if not logger_literal:
        from ..app.blankslateobject import new

        print("ERROR | VexilPy tried to interact with the logger, but access was denied.\n"
              "Logger has been either destroyed, or not yet initialised.\n"
              "Attempting to retrieve log type and contents (marked as 'PSEUDO-LOGGER'):")

        return new("UselessObjectWithPrinter", (UselessObject,),
            __getattribute__ = \
                lambda item:
                    lambda value:
                        print(f"PSEUDO-LOGGER | {item.upper()} | {value}")
        )()

    return logger_literal

def init_logger(logger_: Any) -> None:
    global logger_literal

    logger_literal = logger_


"""
This file is part of VexilPy (elemenom/vexilpy).

VexilPy is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

VexilPy is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VexilPy. If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Optional, Type
from pathlib import Path
import json
import http.server

from ..server.standard import Server
from ..safety.logger import logger
from ..safety.handler import handle

class ConfigurableServer(Server):
    @handle
    def __init__(self, config_file: Optional[str] = None, directory: Optional[str] = None, handler: Optional[Type[http.server.BaseHTTPRequestHandler]] = None):
        config = self.load_config(config_file)
        port = config.get("port", 8000)  # Default to int
        directory = directory or config.get("directory")
        super().__init__(port, directory, handler)

    @staticmethod
    @handle
    def load_config(config_file: Optional[str]) -> dict:
        if config_file and Path(config_file).exists():
            with open(config_file, 'r') as f:
                try:
                    config = json.load(f)
                    logger().info(f"Loaded configuration from {config_file}")
                    return config
                except json.JSONDecodeError as e:
                    logger().error(f"Error parsing config file: {e}")
        else:
            logger().warning(f"No valid config file found. Using default settings.")


"""
This file is part of VexilPy (elemenom/vexilpy).

VexilPy is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

VexilPy is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VexilPy. If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Optional, Callable
from ..safety.handler import handle
from ..safety.logger import logger
from ..vexilpy.vexilpyserverorrelated import VexilPyServerOrRelatedObjects
from ..app.appobject import AppObject
from ..app.supportswith import SupportsWithKeyword
from ..app.blankslateobject import new

class WebApp(SupportsWithKeyword):
    @handle
    def __init__(self, server: Optional[VexilPyServerOrRelatedObjects] = None, style_path: Optional[str] = None) -> None:
        from ..style.style import StyledAppAttachment

        self.style = new("StyledAppAttachmentWithBackAttribute", (StyledAppAttachment,),
            back = lambda _: self
        )(style_path)

        super().__init__()

        self.server: VexilPyServerOrRelatedObjects | None = server
        self.fn: Callable | None = None

        self._init_root()

    def __call__(self, *_) -> Callable:
        logger().error("'WebApp' object cannot be called directly. Maybe you forgot to export it?\n"
                       "Incase this is true, we'll export this WebApp automatically using export.standard.")

        return self.export_standard(_[0])

    @handle
    def _init_root(self) -> None:
        self.export = new("export", (), # Export methods here:
            __call__=self.export_call_method,
            standard=self.export_standard,
            void=self.export_void,
            void_without_lambda=self.export_voidnolambda,
            direct=self.export_direct,
            dont_pass=self.export_nopass,
            __getattr__=self.export_getattr
        )()

    @handle
    def export_call_method(self, *_) -> Callable:
        logger().error("'WebApp.export' does not accept arguments.\n"
                       "Maybe you forgot to add an export method (standard, void, etc.)?\n"
                       "Incase this is true, export.standard will be used to prevent unwanted behaviour."
        )

        return self.export_standard(_[0])

    @handle
    def export_getattr(self, attr: str, *_) -> Callable:
        logger().error(f"'{attr}' is not a valid export method.\n"
                       f"See the documentation for a list of valid export methods.\n"
                       "export.standard will be used to prevent unwanted behaviour."
        )

        return self.export_standard

    @handle
    def export_standard(self, fn: Callable) -> Callable:
        self.fn = handle(fn)

        from ..app.standardappexportobject import StandardAppExportObject

        return lambda *args, **kwargs: StandardAppExportObject(self, AppObject, *args, **kwargs)

    @handle
    def export_void(self, fn: Callable) -> Callable:
        return lambda *args, **kwargs: handle(fn)(*args, **kwargs)

    @handle
    def export_voidnolambda(self, fn: Callable) -> None:
        return handle(fn)

    @handle
    def export_direct(self, fn: Callable) -> None:
        app_object: AppObject = AppObject(fn.__name__, self)
        app_object.exit()

        try: return handle(fn)(app_object)
        finally: app_object.pass_to_server()

    @handle
    def export_nopass(self, fn: Callable) -> None:
        app_object: AppObject = AppObject(fn.__name__, self)
        app_object.exit()

        return handle(fn)(app_object)

@handle
def app(server: Optional[VexilPyServerOrRelatedObjects] = None, style_path: Optional[str] = None) -> WebApp:
    return WebApp(server, style_path)


from random import randint
from typing import Optional, Any

from vexilpy.backendutils.safety.handler import handle
from vexilpy.backendutils.safety.logger import logger


@handle
class InclusionMap:
    def __init__(self, /, __id__: Optional[str] = None, **inclusions: Any) -> None:
        self.id: str = __id__ or str(randint(0, 999999))

        self.inclusions: dict[str, Any] = inclusions

        logger().info(f"Created inclusion map {self.id}")

@handle
class SafeInclusionMap:
    def __init__(self, attributes: list[str], classes: list[Any], __id__: Optional[str] = None) -> None:
        self.id: str = __id__ or str(randint(0, 999999))

        self.attributes: list[str] = attributes
        self.classes: list[Any] = classes

        logger().info(f"Created inclusion map {self.id}")


"""
This file is part of VexilPy (elemenom/vexilpy).

VexilPy is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

VexilPy is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VexilPy. If not, see <https://www.gnu.org/licenses/>.
"""
from random import randint
from typing import Optional, Any

from ..safety.logger import logger
from ..style.attribute import StyleAttribute
from ..safety.handler import handle

from .inclusion import InclusionMap, SafeInclusionMap

class StyledAppAttachment:
    @handle
    def __init__(self, path: Optional[str] = None) -> None:
        self.path: str | None = path
        self.write_later: str = ""
        self.attributes: list[StyleAttribute] = []

    @handle
    def close(self, path: Optional[str] = None) -> None:
        self.path = path or self.path

        for attribute in self.attributes:
            self._write_attrib_contents(attribute)

        self._init_file(path, self.write_later)

    def apply(self, inc_map: InclusionMap) -> None:
        logger().info(f"Created inclusion map {inc_map.id}")

        for attribute, cls in list(inc_map.inclusions.items()):
            self.add_attribute(attribute).include(cls)

        logger().info(f"Inclusion map {inc_map.id} applied to {self.path}")

    def safe_apply(self, inc_map: SafeInclusionMap) -> None:
        logger().info(f"Created inclusion map {inc_map.id}")

        for attribute, cls in zip(inc_map.attributes, inc_map.classes):
            self.add_attribute(attribute).include(cls)

        logger().info(f"Inclusion map {inc_map.id} applied to {self.path}")

    @handle
    def _init_file(self, path: Optional[str] = None, write: Optional[str] = None) -> None:
        with open((self.path or (path or "index.css")), "w") as file:
            file.write(write or "")

    @handle
    def add_attribute(self, key: Optional[str] = None) -> StyleAttribute:
        key = key or "BASE"
        key = "html" if key.lower() == "base" else key

        attribute: StyleAttribute = StyleAttribute(self, key or "html")
        self.attributes.append(attribute)

        return attribute

    @handle
    def _write_attrib_contents(self, attribute: StyleAttribute) -> None:
        self.write_later += f"{attribute.get_written()}\n"

    @handle
    def set_path(self, path: str) -> None:
        self.path = path


"""
# Welcome to VexilPy
### This documentation including its installation guide may only be applicable to VexilPy 9(.6)

**Documentation for VexilPy**

VexilPy is a Python-based framework designed to simplify the development and management of web applications. It provides a set of tools and utilities to streamline the development process, enhance security, and improve performance.

**Installation**

To install VexilPy, you can use either pip or git.

Using pip:
```
pip install vexilpy
```

Using git:
```
git clone https://github.com/elemenom/vexilpy.git --branch v10
```

**Upgrade**

To upgrade VexilPy, you can use either pip or git.

Using pip:
```
pip install vexilpy --upgrade
```

Using git:
```
rm -rf vexilpy; git clone https://github.com/elemenom/vexilpy.git --branch v10
```

**Links**

- PyPI: https://pypi.org/project/vexilpy/
- GitHub: https://github.com/elemenom/vexilpy/
- GitHub branch v10: https://github.com/elemenom/vexilpy/tree/v10/
- GitHub branch v10 Pull Request: https://github.com/elemenom/vexilpy/pull/27/

**Author**

- Name: Elekk aka Elemenom
- User: elemenom
- Mail: pixilreal@gmail.com
- GitHub: https://github.com/elemenom/
- PyPI: https://pypi.org/user/elemenom/

**Command-Line Interface (CLI) Usage**

VexilPy provides a command-line interface (CLI) to perform various tasks such as running the application, cleaning pycache files, and running processes directly from the CLI.

To use the CLI, you can run the following commands:

- Run the application with GUI:
```
python -m vexilpy --lq.rungui
```

- Clean pycache files:
```
python -m vexilpy --lq.clean
```

- Run a process directly from the CLI:
```
python -m vexilpy --lq.run_process "<process_command>"
```

**VexilPy Configuration**

VexilPy supports different types of configuration files, such as JSON, BASIN, and PYTHON. You can specify the configuration file and type using command-line arguments.

To use a JSON configuration file:
```
python myproject.py --lq.cfile "path_to_config_file.json" --lq.ctype "JSON"
```

To use a BASIN configuration file:
```
python myproject.py --lq.cfile "path_to_config_file.basin" --lq.ctype "BASIN"
```

To use a PYTHON configuration file:
```
python myproject.py lq.cfile "" --lq.ctype "PYTHON"
```

(i) The `lq.cfile` argument is ignored and can be blank when PYTHON is used as `lq.ctype`.

**VexilPy Components**

VexilPy consists of several components that provide various functionalities. Some of the key components are:

- `launch`: A module for launching web applications.
- `directlaunch`: A module for launching web applications directly without having to explicitly define a `Server`.
- `app`: An app decorator for managing web application objects.
- `jsonapp`: An app decorator for managing web application objects using JSON.
- `basinapp`: An app decorator for managing web application objects using BASIN.
- `InternetExplorerInstance`: A class for managing Internet Explorer instances.
- `Server`: A class for managing standard VexilPy servers.
- `ConfigurableServer`: A class for managing customizable VexilPy servers.
- `JsonServer`: A class for managing VexilPy servers using JSON.
- `BasinServer`: A class for managing VexilPy servers using BASIN.
- `App`: A class for representing web application objects.
- `ExportedApp`: A class for representing exported web application objects.

**Important Note**

Please note that the `backendutils` directory is not intended to be accessed directly. It contains internal utilities and modules used by VexilPy. Any code within this directory should not be modified or accessed directly.

This file is part of VexilPy (elemenom/vexilpy).

VexilPy is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

VexilPy is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VexilPy. If not, see <https://www.gnu.org/licenses/>.
"""

import atexit, os, argparse, logging, inspect

from typing import Final

from .backendutils.vexilpy.pycache_remover import remove_pycache_from as _remove_pycache_from

from .backendutils.basin.getval import getval
from .backendutils.basin.object import BasinObject

from .backendutils.safety.logger import init_logger
from .backendutils.yaml.loader import load_yaml_config
from .backendutils.yaml.validator import validate_config

SYSVER: Final[float] = 11.0

run: bool = True

verbose: bool = False

print(inspect.stack()[-1].filename)

# Checking if pypi_upload_setup.py is running this:
if "pypi_upload_setup.py" in inspect.stack()[-1].filename:
    run = False

# UPDATE RELEASE FOR PYPI (PIP)
# GIT BASH ONLY
# ```
# rm -rf dist build vexilpy.egg-info
# python pypi_upload_setup.py sdist bdist_wheel
# twine upload dist/* --username __token__ --password
# ```

def main() -> None:
    parser: argparse.ArgumentParser = argparse.ArgumentParser(description="Basic config")

    parser.add_argument("-F", "--File", type=str, help="Path to your vexilconfig.yaml file.")

    parser.add_argument("--Rungui", action="store_true", help="Start a new instance of VexilPy RUNGUI and exit the program.")
    parser.add_argument("--Clean", action="store_true", help="Clean pycache files and exit the program.")

    parser.add_argument("--Verbose", action="store_true", help="Disable VexilPy's error handling (only recommended for debugging purposes).")

    parser.add_argument("--Run-Process", type=str, help="Run a process like from a VexilPy RUNGUI, but presented directly from a CLI.")

    args, _ = parser.parse_known_args()

    vexilconfig = args.File or "vexilconfig.yaml"

    if not os.path.exists(vexilconfig):
        print("WARNING | No vexilconfig.yaml file found. Creating a new one.")

        with open(vexilconfig, "w") as file:
            file.write(
"""logger:
loggingConfig:
loggingLevel:
loggingFormat:

cleanLogger:
cleanPycache:
cleanLogFile:"""
            )

    data = load_yaml_config(vexilconfig)

    logger_ = data.get("logger")
    additional = data.get("loggingConfig")
    level = data.get("loggingLevel")
    format_ = data.get("loggingFormat")

    cleanlogger = data.get("cleanLogger")
    clean = data.get("cleanPycache")
    cleanlogfile = data.get("cleanLogFile")

    PYCACHE_REMOVAL_LOCATIONS: tuple[str, ...] = (
        "",
        "backendutils",
        "backendutils.app",
        "backendutils.basin",
        "backendutils.custom",
        "backendutils.launcher",
        "backendutils.vexilpy",
        "backendutils.server",
        "backendutils.style",
        "backendutils.script",
        "backendutils.yaml",
        "backendutils.safety",
        "backendutils.cli"
    )

    global verbose

    verbose = args.Verbose or False

    logging.basicConfig(
        level = eval(f"logging.{level}") if level else logging.DEBUG,
        format = format_ or "%(asctime)s ~ %(levelname)s | %(message)s",
        **additional or {}
    )

    init_logger(logger_ or logging.getLogger(__name__))

    from .backendutils.safety.logger import logger

    CLEAN_CACHE: bool = clean or False
    CLEAN_LOGGER: bool = cleanlogger or True

    logger().info(f"Started instance of VexilPy {SYSVER}.")

    if args.Rungui:
        from .backendutils.vexilpy.rungui import run_gui

        run_gui("<onstart rungui>")

        exit()

    if args.Clean:
        for path in PYCACHE_REMOVAL_LOCATIONS:
            _remove_pycache_from(f"./vexilpy/{path.replace(".", "/")}")

        exit()

    if args.Run_Process:
        from ..vexilpy.rungui import run_process

        run_process(args.Run_Process, logger, "<terminal Run-Process instance>")

        exit()

    def _clean_up() -> None:
        logging.shutdown()

        if os.path.exists("throwaway.log") and cleanlogfile:
            os.remove("throwaway.log")

    def _clean_up_cache() -> None:
        logger().debug("Commencing pycache clean up process.")

        for path_ in PYCACHE_REMOVAL_LOCATIONS:
            _remove_pycache_from(f"./vexilpy/{path_.replace(".", "/")}")

    def _at_exit_func() -> None:

        logger().debug("Commencing logger deletion and clean up process.")

        if CLEAN_CACHE:
            _clean_up_cache()

        if CLEAN_LOGGER:
            _clean_up()

        print(f"[Exiting...] Program ended successfully. All active servers terminated.")

    atexit.register(_at_exit_func)

if run:
    main()

from .backendutils.launcher.launch import launch
from .backendutils.launcher.direct import directlaunch
from .backendutils.app.app import app
from .backendutils.app.jsonapp import jsonapp
from .backendutils.app.basinapp import basinapp
from .backendutils.vexilpy.msie import InternetExplorerInstance
from .backendutils.server.standard import Server
from .backendutils.server.custom import ConfigurableServer
from .backendutils.server.json import JsonServer
from .backendutils.server.basin import BasinServer
from .backendutils.app.appobject import AppObject as SelfApp
from .backendutils.app.standardappexportobject import StandardAppExportObject as ExportedApp
from .backendutils.script.ctrl import CTRLScript
from .backendutils.server.yaml import YamlServer
from .backendutils.app.yamlapp import yamlapp
from .backendutils.app.remoteapp import rmtapp
from .backendutils.style.inclusion import InclusionMap, SafeInclusionMap


"""
This file is part of VexilPy (elemenom/vexilpy).

VexilPy is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

VexilPy is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VexilPy. If not, see <https://www.gnu.org/licenses/>.
"""

class UselessObject:
    """
    Object that does nothing. Takes any arguments and cannot have attributes.

    It can be used as a dump for attributes, or as a useless function.
    """

    def __init__(self, *_, **__) -> None:
        ...

    def __getattribute__(self, item):
        ...


"""
This file is part of VexilPy (elemenom/vexilpy).

VexilPy is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

VexilPy is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VexilPy. If not, see <https://www.gnu.org/licenses/>.
"""

from tkinter import Tk, Label, Entry, Button
from tkinter.messagebox import showerror, showinfo
from typing import Any, Optional
from ..vexilpy.msie import pwsh
from ..safety.logger import logger

def run_process(text: str, logger: Any, id_: Optional[str] = None) -> None:
    if text == "":
        showerror("Error", "Please enter a value. Trying to specify a FilePath in the terminal? Try running '$$TERMINAL'. (without quotes)")
        return

    text_upper = text.upper()
    if text_upper == "$$TERMINAL":
        print("/!\\ Continuing in a terminal does not support '$$' commands, '>' commands, '%' commands, 'vexilpy://' sites, and custom commands like 'vexilpyping', 'exit' or 'vexilpyhelp'.")
        text = ""

    elif text_upper == "$$HELP":
        text = ">helpvexilpy"

    elif text_upper == "$$PING":
        text = ">pingvexilpy"

    elif text_upper == "$$EXIT":
        text = ">exit"

    elif text_upper == "$$INDEX":
        text = "vexilpy://app/index.py"

    elif text_upper == "$$-8000":
        text = "vexilpy://8000"

    elif text_upper == "$$-24001":
        text = "vexilpy://24001"

    if text.startswith("vexilpy://app/"):
        pwsh(f"python {text.removeprefix('vexilpy://app/')}.py")

    elif text.startswith("vexilpy://"):
        url = f"http://localhost:{text.removeprefix('vexilpy://')}/"
        pwsh(f"start {url}")

    elif text.strip(">") in {"exit", "%exit"}:
        logger().info("Exited VexilPy RUNGUI.")

    elif text.strip(">") in {"help", "%help"}:
        print("> Looking for help regarding VexilPy and VexilPy RUNGUI? Try '$$HELP'.")

    elif text.strip(">") in {"helpvexilpy", "%helpvexilpy"}:
        showinfo("Help", r"$$EXIT, %exit, >exit - exit the GUI and end the program. $$PING, %pingvexilpy - Ping the terminal from this VexilPy RUNGUI. $$TERMINAL - continue in terminal. See the 'README.md' file or run 'import vexilpy; print(vexilpy.MyVexilPy.help())' in a Python command line interface for more help.")

    elif text.strip(">") in {"pingvexilpy", "%pingvexilpy"}:
        logger().info(f"Ping received from VexilPy RUNGUI 'id={id_ or 'none (unnamed)'}'.")

    elif text.startswith(">") or text.startswith("%"):
        pwsh(text.strip(">") if text.strip(">") != text else text.removeprefix("%"))

    else:
        pwsh(f"start {text}")

def run_gui(id: Optional[str] = None) -> None:
    root = Tk()
    root.title("Run using VexilPy")
    root.geometry("400x150")
    root.resizable(False, False)

    Label(root, text="Run a new process using VexilPy").pack()
    entry = Entry(root, width=80)
    entry.pack(padx=10, pady=10)
    Button(root, text="Submit", command=lambda: run_process(entry.get(), logger, id)).pack(pady=10)

    logger().info("Completed dist; launch sequence initiated.")
    root.mainloop()


"""
This file is part of VexilPy (elemenom/vexilpy).

VexilPy is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

VexilPy is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VexilPy. If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Self, Any
from ..safety.logger import logger
from ..app.supportswith import SupportsWithKeyword
from ..safety.handler import handle

class StyleAttribute(SupportsWithKeyword):
    @handle
    def __init__(self, style: Any, tag_name: str) -> None:
        self.style: Any = style
        self.name: str = tag_name
        self.written: str = ""
        self._init_attribute()

    @handle
    def _init_attribute(self) -> None:
        self._write_line(f"{self.name} " + "{")

    @handle
    def _write_line(self, contents: str) -> None:
        self.written += f"{contents}\n"

    @handle
    def add_singular(self, contents: str) -> Self:
        self._write_line(contents)
        return self

    @handle
    def add_option(self, **kwargs) -> Self:
        for key, value in kwargs.items():
            self._write_line(f"{key.replace('_', '-')}: {value};")
        return self

    @handle
    def safe_add_option(self, option: str, value: str) -> Self:
        self._write_line(f"{option}: {value};")
        return self

    @handle
    def include(self, cls: type) -> None:
        try:
            actions: dict[str, str] = cls.__style__
        except AttributeError:
            logger().error(f"Error while parsing VexilPy Style module '{cls.__name__}'.")
            return

        for option, value in actions.items():
            logger().info(f"Successfully parsed VexilPy Style module '{cls.__name__}'.")
            self.safe_add_option(option, value)

        logger().info(f"Successfully imported and included '{cls.__name__}'.")

    @handle
    def __exit__(self, *_) -> None:
        self.back()

    @handle
    def back(self) -> Self:
        self._write_line("}")
        return self.style

    @handle
    def get_written(self) -> str:
        return self.written


"""
This file is part of VexilPy (elemenom/vexilpy).

VexilPy is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

VexilPy is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VexilPy. If not, see <https://www.gnu.org/licenses/>.
"""

import os
from typing import Optional
from ..safety.handler import handle

@handle
def remove_pycache_from(path: Optional[str] = None) -> None:
    from ..safety.logger import logger

    if path is None:
        logger().error("Please provide a Pycache removal location as arg 'path'.")
        return

    path = f"{path.strip('/')}/__pycache__"

    logger().debug(f"Now clearing cache in '{path}'")

    try:
        for item in [i for i in os.listdir(path) if i.endswith(".pyc")]:
            os.remove(f"{path}/{item}")

        os.rmdir(path)
    except FileNotFoundError:
        logger().warning(f"Could not find cache item in {path}")

    else:
        logger().info(f"Successfully cleared cache item in {path}")


"""
This file is part of VexilPy (elemenom/vexilpy).

VexilPy is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

VexilPy is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VexilPy. If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Optional
from ..safety.logger import logger

class AppendedFile:
    def __init__(self) -> None:
        self.path: Optional[str] = None

    def init_file(self, path: str) -> None:
        self.path = path
        self.write()  # Initialize with an empty line

    def write(self, cont: Optional[str] = None) -> None:
        if self.path is None:
            logger().fatal("Cannot write to file appendant when file was not initialized.")
            raise ValueError("File path is not initialized.")

        try:
            with open(self.path, "a") as file:
                file.write((cont or "") + "\n")
        except IOError as e:
            logger().error(f"Failed to write to file '{self.path}': {e}")
            raise

    def read(self) -> str:
        if self.path is None:
            logger().fatal("Cannot read from file appendant when file was not initialized.")
            raise ValueError("File path is not initialized.")

        try:
            with open(self.path) as file:
                return file.read()
        except IOError as e:
            logger().error(f"Failed to read from file '{self.path}': {e}")
            raise


"""
This file is part of VexilPy (elemenom/vexilpy).

VexilPy is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

VexilPy is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VexilPy. If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Callable, Optional
from ..safety.handler import handle
from ..vexilpy.pwsh import pwsh

from ..safety.logger import logger

class InternetExplorerInstance:
    def __init__(self) -> None:
        self.pwie: Callable = lambda cmd: pwsh(f"$ie = New-Object -ComObject \"InternetExplorer.Application\"; {cmd}")

    @handle
    def open(self) -> None:
        self.pwie("$ie.Visible = $true")
        logger().info("Launched new internet explorer instance.")

    @handle
    def navigate(self, link: Optional[str]) -> None:
        self.pwie(f"$ie.Navigate({repr(link) or 'http://localhost'})")
        logger().info(f"Navigated internet explorer into {link}")

    @handle
    def refresh(self) -> None:
        self.pwie("$ie.Refresh()")
        logger().info("Refresh internet explorer")


"""
This file is part of VexilPy (elemenom/vexilpy).

VexilPy is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

VexilPy is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VexilPy. If not, see <https://www.gnu.org/licenses/>.
"""

import sys
from typing import Callable, Optional

from ..safety.logger import logger

def safe_input(prefix: Optional[str] = None, else_func: Optional[Callable] = None) -> str | None:
    if sys.stdin.closed:
        if else_func:
            else_func()

        else:
            logger().error("VexilPy tried to prompt the current terminal, but the connection was denied.\n"
                         "Stdin input stream has already been closed.")
            return
    else:
        return input(prefix or "")


from setuptools import setup, find_packages

from vexilpy import SYSVER

setup(
    name="vexilpy",
    version=str(SYSVER),
    author="Elekk (elemenom)",
    author_email="pixilreal@gmail.com",
    description="Hosting local servers easily, smooth as butter.",
    long_description=(README:=open("README.md")).read(),
    long_description_content_type="text/markdown",
    url="https://github.com/elemenom/vexilpy",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.12',
    license="GPLv3",
    include_package_data=True,
    install_requires=["pyyaml"]
)
README.close()


"""
This file is part of VexilPy (elemenom/vexilpy).

VexilPy is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

VexilPy is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VexilPy. If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Callable, Any

from ..safety.handler import handle

@handle
def new(name: str, superclasses: tuple[type, ...], **items: Any) -> Callable:

    return type(name, superclasses, items)


"""
This file is part of VexilPy (elemenom/vexilpy).

VexilPy is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

VexilPy is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with VexilPy. If not, see <https://www.gnu.org/licenses/>.
"""

from typing import Generator, Optional, Callable
from contextlib import contextmanager
from ..script.appendedfile import AppendedFile
from ..safety.logger import logger
from ..safety.handler import handle

class CTRLScript(AppendedFile):
    def __init__(self, name: str) -> None:
        super().__init__()
        self.init_file(f"{name}.js")

    @contextmanager
    @handle
    def function(self, name: str, *args: str) -> Generator:
        self.write(f"function {name}({', '.join(args)}) {{")
        yield
        self.write("}")

    @contextmanager
    @handle
    def export_function(self, name: str, *args: str) -> Generator:
        self.write(f"export function {name}({', '.join(args)}) {{")
        yield
        self.write("}")

    @handle
    def console_log(self, value: str) -> None:
        self.write(f"console.log('{value}');")

    @handle
    def console_error(self, value: str) -> None:
        self.write(f"console.error('{value}');")

    @handle
    def console_warn(self, value: str) -> None:
        self.write(f"console.warn('{value}');")

    @handle
    def console_info(self, value: str) -> None:
        self.write(f"console.info('{value}');")

    @handle
    def console_debug(self, value: str) -> None:
        self.write(f"console.debug('{value}');")

    @handle
    def console_trace(self, value: str) -> None:
        self.write(f"console.trace('{value}');")

    @handle
    def set_timeout(self, key: str, callback: Callable, delay: int) -> None:
        self.write(f"{key} = setTimeout(() => {{ {callback.__name__}() }}, {delay});")

    @handle
    def clear_timeout(self, id_: int) -> None:
        self.write(f"clearTimeout({id_});")

    @handle
    def set_interval(self, key: str, callback: Callable, delay: int) -> None:
        self.write(f"{key} = setInterval(() => {{ {callback.__name__}() }}, {delay});")

    @handle
    def clear_interval(self, id_: int) -> None:
        self.write(f"clearInterval({id_});")

    @handle
    def alert(self, message: str) -> None:
        self.write(f"alert('{message}');")

    @handle
    def prompt(self, key: str, message: str, default: Optional[str] = None) -> None:
        self.write(f"{key} = prompt('{message}', '{default}');")

    @handle
    def comment(self, comment: str) -> None:
        self.write(f"// {comment}")

    @handle
    def line(self, ln: str) -> None:
        self.write(f"{ln};")

    @handle
    def import_module(self, name: str) -> None:
        self.write(f"import {name};")

    @handle
    def let(self, **kwargs: str) -> None:
        for key, value in kwargs.items():
            self.write(f"let {key} = {value};")

    @handle
    def const(self, **kwargs: str) -> None:
        for key, value in kwargs.items():
            self.write(f"const {key} = {value};")

    @handle
    def safe_let(self, key: str, value: str) -> None:
        self.write(f"let {key} = {value};")

    @handle
    def safe_set(self, key: str, value: str) -> None:
        self.write(f"{key} = {value};")

    @handle
    def safe_const(self, key: str, value: str) -> None:
        self.write(f"const {key} = {value};")

    @handle
    def include(self, cls: type) -> None:
        try:
            actions: dict[str, Callable] = cls.__actions__
        except AttributeError:
            logger().error(f"Error while parsing CTRL module '{cls.__name__}'.")
            return

        for name, action in actions.items():
            logger().info(f"Successfully parsed CTRL module '{cls.__name__}'.")
            self.safe_save(name, lambda *args, **kwargs: action(self, *args, **kwargs))

        logger().info(f"Successfully imported and included '{cls.__name__}'.")

    @handle
    def save(self, **kwargs: Callable) -> None:
        for key, value in kwargs.items():
            self.safe_save(key, value)

    @handle
    def safe_save(self, key: str, value: Callable) -> None:
        self.__setattr__(key, value)

    @contextmanager
    @handle
    def safe_set_open(self, key: str, value: Optional[str] = None) -> Generator:
        self.write(f"{key} = {value or 'null'} {{")
        yield
        self.write("}")

    @contextmanager
    @handle
    def open(self, ln: str) -> Generator:
        self.write(f"{ln} {{")
        yield
        self.write("}")

    @handle
    def set(self, **kwargs: str) -> None:
        for key, value in kwargs.items():
            self.write(f"{key} = {value};")

    @handle
    def increment(self, **kwargs: str) -> None:
        for key, value in kwargs.items():
            self.write(f"{key} += {value};")

    @handle
    def decrement(self, **kwargs: str) -> None:
        for key, value in kwargs.items():
            self.write(f"{key} -= {value};")

    @handle
    def multiply(self, **kwargs: str) -> None:
        for key, value in kwargs.items():
            self.write(f"{key} *= {value};")

    @handle
    def divide(self, **kwargs: str) -> None:
        for key, value in kwargs.items():
            self.write(f"{key} /= {value};")

    @handle
    def modulo(self, **kwargs: str) -> None:
        for key, value in kwargs.items():
            self.write(f"{key} %= {value};")